unit SuggestEdit;

{$mode objfpc}{$H+}

interface

uses
  Classes, sysutils, StdCtrls, Controls, LCLType, LCLProc,
  Graphics, LCLIntf, LMessages, LazUTF8, SuggestEditListbox;

type

  TStringEvent = procedure(Sender: TObject; const AText: String) of object;

  TCompareOption = (coCaseSensitive, coAnywhere);
  TCompareOptions = Set of TCompareOption;

  { TCustomSuggestEdit }

  TCustomSuggestEdit = class(TVirtualCustomSuggestEdit)
  private
    FBox: TSuggestEditListboxForm;
    FKeyForListBox: Boolean;
    FOldText: TCaption;
    FOnDeleteItem: TStringEvent;
    function GetCompareOptions: TCompareOptions;
    function GetItems: TStringList;
    function GetMaxBoxHeight: Integer;
    function GetMaxPreviewCount: Integer;
    procedure SelectItem(const s: String);
    procedure SetCompareOptions(AValue: TCompareOptions);
    procedure SetMaxBoxHeight(AValue: Integer);
    procedure SetMaxPreviewCount(AValue: Integer);
  protected
    procedure Change; override;
    procedure DeleteItem;
    procedure DoFocus; override;
    procedure KeyDown(var Key: Word; Shift: TShiftState); override;
    procedure SetColor(Value: TColor); override;
    procedure SetParent(NewParent: TWinControl); override;
    procedure WndProc(var Message: TLMessage); override;
  protected
    property CompareOptions: TCompareOptions read GetCompareOptions write SetCompareOptions default [coAnywhere];
    property Items: TStringList read GetItems;
    property MaxBoxHeight: Integer read GetMaxBoxHeight write SetMaxBoxHeight default 200;
    property MaxBoxPreviewCount: Integer read GetMaxPreviewCount write SetMaxPreviewCount default -1;
    property OnDeleteItem: TStringEvent read FOnDeleteItem write FOnDeleteItem;
  public
    constructor Create(AOwner: TComponent); override;
    procedure EditingDone; override;
    function Focused: Boolean; override;
    procedure SelectionChange(const aValue: string);
    procedure SetBounds(ALeft, ATop, AWidth, {%H-}AHeight: Integer); override;
  public
    property AutoSelected default False;
    property AutoSize default True;
    property BorderStyle default bsSingle;
    property CaretPos: TPoint read GetCaretPos write SetCaretPos;
    property NumbersOnly: Boolean read GetNumbersOnly write SetNumbersOnly default false;
    property ReadOnly: Boolean read GetReadOnly write SetReadOnly default false;
    property SelLength: integer read GetSelLength write SetSelLength;
    property SelStart: integer read GetSelStart write SetSelStart;
    property SelText: String read GetSelText write SetSelText;
    property TabStop default true;
  end;

  TSuggestEdit = class(TCustomSuggestEdit)
  published
    property Action;
    property Align;
    property Alignment;
    property Anchors;
    property AutoSize;
    property AutoSelect;
    property BidiMode;
    property BorderSpacing;
    property BorderStyle;
    property CharCase;
    property Color;
    property CompareOptions;
    property Constraints;
    property Cursor;
    property DragCursor;
    property DragKind;
    property DragMode;
    property EchoMode;
    property Enabled;
    property Font;
    property HideSelection;
    property Items;
    property MaxBoxHeight;
    property MaxBoxPreviewCount;
    property MaxLength;
    property NumbersOnly;
    property ParentBidiMode;
    property OnChange;
    property OnChangeBounds;
    property OnClick;
    property OnContextPopup;
    property OnDblClick;
    property OnDeleteItem;
    property OnDragDrop;
    property OnDragOver;
    property OnEditingDone;
    property OnEndDrag;
    property OnEnter;
    property OnExit;
    property OnKeyDown;
    property OnKeyPress;
    property OnKeyUp;
    property OnMouseDown;
    property OnMouseEnter;
    property OnMouseLeave;
    property OnMouseMove;
    property OnMouseUp;
    property OnMouseWheel;
    property OnMouseWheelDown;
    property OnMouseWheelUp;
    property OnResize;
    property OnStartDrag;
    property OnUTF8KeyPress;
    property ParentColor;
    property ParentFont;
    property ParentShowHint;
    property PasswordChar;
    property PopupMenu;
    property ReadOnly;
    property ShowHint;
    property TabStop;
    property TabOrder;
    property Text;
    property TextHint;
    property TextHintFontColor;
    property TextHintFontStyle;
    property Visible;
  end;

implementation

{ TCustomSuggestEdit }

function TCustomSuggestEdit.GetCompareOptions: TCompareOptions;
begin
  Result := TCompareOptions(FBox.CompareOptions);
end;

function TCustomSuggestEdit.GetItems: TStringList;
begin
  Result := FBox.Items;
end;

function TCustomSuggestEdit.GetMaxBoxHeight: Integer;
begin
  Result := FBox.MaxBoxHeight
end;

function TCustomSuggestEdit.GetMaxPreviewCount: Integer;
begin
  Result := FBox.MaxPreviewCount;
end;

procedure TCustomSuggestEdit.SelectItem(const s: String);
begin
  Text := s;
  FOldText := s;
  FBox.Visible := False;
end;

procedure TCustomSuggestEdit.SetCompareOptions(
  AValue: TCompareOptions);
begin
  FBox.CompareOptions := SuggestEditListbox.TCompareOptions(AValue);
end;

procedure TCustomSuggestEdit.SetMaxBoxHeight(AValue: Integer);
begin
  FBox.MaxBoxHeight := AValue;
end;

procedure TCustomSuggestEdit.SetMaxPreviewCount(AValue: Integer);
begin
  FBox.MaxPreviewCount := AValue;
end;

procedure TCustomSuggestEdit.Change;
begin
  inherited Change;
  if not FKeyForListBox then
  begin
    FOldText := Text;
    FBox.FillBox(Text);
  end;
end;

procedure TCustomSuggestEdit.DeleteItem;
begin
  if Assigned(FOnDeleteItem) then begin
    FKeyForListBox := True;
    FOnDeleteItem(Self, FBox.SelectedItem);
    FBox.DeleteIndex;
  end;
end;

procedure TCustomSuggestEdit.DoFocus;
begin
  SetFocus;
  SelStart := Length(Text);
end;

procedure TCustomSuggestEdit.KeyDown(var Key: Word;
  Shift: TShiftState);
begin
//  writeln('TCustomSuggestEdit.KeyDown ', Key, ' - ', IntToHex(Key, 4));
  case Key of
    VK_UP:
      begin
        FKeyForListBox := True;
        if FBox.Showing then FBox.MoveItemIndexUp;
        Key := 0;
      end;
    VK_DOWN:
      begin
        FKeyForListBox := True;
        if FBox.Showing then FBox.MoveItemIndexDown
                        else FBox.FillBox(Text);
        Key := 0;
      end;
    VK_PRIOR: // Page Up
      begin
        FKeyForListBox := True;
        if FBox.Showing then FBox.MoveItemIndexPageUp;
        Key := 0;
      end;
    VK_NEXT: // Page Down
      begin
        FKeyForListBox := True;
        if FBox.Showing then FBox.MoveItemIndexPageDown
                        else FBox.FillBox(Text);
        Key := 0;
      end;
    VK_RETURN:
      begin
        if FBox.ItemSelected and FKeyForListBox then
          SelectItem(FBox.SelectedItem)
        else
          FBox.Visible := False;
      end;
    VK_TAB:
      begin
        if FBox.Showing and FBox.ItemSelected then
        begin
          Text := FBox.SelectedItem;
          FOldText := FBox.SelectedItem;
        end;
        FBox.Visible := False;
      end;
    VK_ESCAPE:
      begin
        if FBox.Showing then
        begin
          FKeyForListBox := True;
          FBox.Visible := False;
          Text := FOldText;
          SelStart := Length(FOldText);
        end;
        Key := 0;
      end;
    VK_DELETE:
      if FBox.Visible and FBox.ItemSelected then
        DeleteItem;
    else
      FKeyForListBox := False;
  end;
  inherited KeyDown(Key, Shift);
end;

procedure TCustomSuggestEdit.SetColor(Value: TColor);
begin
  inherited SetColor(Value);
  FBox.Color := Value;
end;

procedure TCustomSuggestEdit.SetParent(NewParent: TWinControl);
begin
  inherited SetParent(NewParent);
  FBox.Parent := Nil;
end;

procedure TCustomSuggestEdit.WndProc(var Message: TLMessage);
const
  {$IFDEF WINDOWS}
  WM_IME_NOTIFY = $0282;
  {$ENDIF}
  CP: TPoint = (x: 0; y: 0);
var
  p: TPoint;
begin
  if {$IFDEF WINDOWS} (Message.msg = WM_IME_NOTIFY) {$ELSE} (Message.msg = LM_PAINT) {$ENDIF} then begin
    p := ClientToScreen(Point(0,0));
    if (p.x <> CP.x) or (p.y <> CP.y) then begin
      CP := p;
      FBox.Visible := False;
    end;
  end;
  writeln('TCustomSuggestEdit.WndProc ', Message.msg, ' - ', IntToHex(Message.Msg, 4));
  inherited WndProc(Message);
end;

constructor TCustomSuggestEdit.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FBox := TSuggestEditListboxForm.Create(Self);
  FBox.OnSelectionChange := @SelectionChange;
  FBox.OnSelectItem := @SelectItem;
end;

procedure TCustomSuggestEdit.EditingDone;
begin
  if not FBox.Visible then
    inherited EditingDone;
end;

function TCustomSuggestEdit.Focused: Boolean;
begin
  Result := inherited Focused;
  if Assigned(FBox) then Result := Result or FBox.Focused;
end;

procedure TCustomSuggestEdit.SelectionChange(const aValue: string);
var
  s: string;
begin
  FKeyForListBox := True;
  if FBox.ItemSelected then s := aValue
                       else s := FOldText;
  Text := s;
  SelStart := Length(s);
end;

procedure TCustomSuggestEdit.SetBounds(ALeft, ATop, AWidth, AHeight: Integer);
begin
  inherited SetBounds(ALeft, ATop, AWidth, Height);
  if Assigned(FBox) then
    FBox.ResizeForm;
end;

end.

