unit SuggestEditListbox;

{$mode objfpc}{$H+}

interface

uses
  Classes, SysUtils, math, Types, Forms, StdCtrls, Controls,
  LCLType, LCLIntf, Graphics;

type

  TCompareOption = (coCaseSensitive, coAnywhere);
  TCompareOptions = Set of TCompareOption;

  TVirtualCustomSuggestEdit = class(TCustomEdit)
  protected
    procedure DoFocus; virtual; abstract;
  end;

  { TSuggestEditListboxForm }

  TSuggestEditListboxForm = class(TForm)
  private
    FBorderHeight: Integer;
    FCompareOptions: TCompareOptions;
    FCurrentEditText: String;
    FFirstShow: Boolean;
    FItemHeight: Integer;
    FItemsDummy: TStringList;
    FItemShowingCount: Integer;
    FListBox: TListBox;
    FMaxBoxHeight: Integer;
    FMaxPreviewCount: Integer;
    FOnSelectionChange: TGetStrProc;
    FOnSelectItem: TGetStrProc;
    FSuggestEdit: TVirtualCustomSuggestEdit;
    procedure ClickItem({%H-}Sender: TObject);
    procedure FillItemsDummy;
    procedure FocusEdit({%H-}Sender: TObject);
    function GetItemSelected: Boolean;
    function GetSelectedItem: String;
    procedure MouseMoveItems({%H-}Sender: TObject; {%H-}Shift: TShiftState; X, Y: Integer);
    procedure PrepareFirstShow;
    procedure SelectItem;
  protected
    procedure SetColor(Value: TColor); override;
  public
    Items: TStringList;
    constructor Create(AOwner: TComponent); override;
    constructor CreateNew(AOwner: TComponent; Num: Integer = 0); override;
    destructor Destroy; override;
    procedure DeleteIndex;
    procedure FillBox(const s: String);
    procedure FocusOwner({%H-}Data: PtrInt);
    function GetSelectedText: String;
    procedure MoveItemIndexDown;
    procedure MoveItemIndexPageDown;
    procedure MoveItemIndexPageUp;
    procedure MoveItemIndexUp;
    procedure SelectionChanged({%H-}Sender: TObject; {%H-}User: boolean);
    procedure ResizeForm;
  public
    property CompareOptions: TCompareOptions read FCompareOptions write FCompareOptions;
    property ListBox: TListBox read FListBox;
    property MaxBoxHeight: Integer read FMaxBoxHeight write FMaxBoxHeight;
    property MaxPreviewCount: Integer read FMaxPreviewCount write FMaxPreviewCount default -1;
    property OnSelectionChange: TGetStrProc read FOnSelectionChange write FOnSelectionChange;
    property OnSelectItem: TGetStrProc read FOnSelectItem write FOnSelectItem;
    property ItemSelected: Boolean read GetItemSelected;
    property SelectedItem: String read GetSelectedItem;
  end;

implementation

{ TSuggestEditListboxForm }

procedure TSuggestEditListboxForm.ClickItem(Sender: TObject);
begin
  SelectItem;
end;

procedure TSuggestEditListboxForm.FillItemsDummy;
var
  i: Integer;
begin
  FItemsDummy.Clear;
  if Length(FCurrentEditText) = 0 then begin
    if MaxPreviewCount = -1 then
      FItemsDummy.Assign(Items)
    else
      for i := 0 to Min(FMaxPreviewCount, Items.Count) -1 do
        FItemsDummy.Add(Items[i]);
    Exit;
  end;

  for i := 0 to Pred(Items.Count) do
  begin
    if coCaseSensitive in CompareOptions then
      if coAnywhere in CompareOptions then
      begin
        if Pos(FCurrentEditText, Items[i]) > 0 then
          FItemsDummy.Add(Items[i]);
      end
      else
      begin
        if Pos(FCurrentEditText, Items[i]) = 1 then
          FItemsDummy.Add(Items[i]);
      end
    else
      if coAnywhere in CompareOptions then
      begin
        if Pos(LowerCase(FCurrentEditText), LowerCase(Items[i])) > 0 then
          FItemsDummy.Add(Items[i]);
      end
      else
      begin
        if Pos(LowerCase(FCurrentEditText), LowerCase(Items[i])) = 1 then
          FItemsDummy.Add(Items[i]);
      end;
    if FMaxPreviewCount = -1 then Continue;
    if FItemsDummy.Count = FMaxPreviewCount then Break;
  end;
end;

procedure TSuggestEditListboxForm.FocusEdit(Sender: TObject);
begin
  Application.QueueAsyncCall(@FocusOwner, 0);
end;

function TSuggestEditListboxForm.GetItemSelected: Boolean;
begin
  Result := FListBox.ItemIndex >= 0;
end;

function TSuggestEditListboxForm.GetSelectedItem: String;
begin
  if FListBox.ItemIndex >= 0 then
    Result := FListBox.Items[FListBox.ItemIndex]
  else
    Result := '';
end;

procedure TSuggestEditListboxForm.MouseMoveItems(
  Sender: TObject; Shift: TShiftState; X, Y: Integer);
begin
  FListBox.ItemIndex := FListBox.ItemAtPos(Point(X, Y), True);
end;

procedure TSuggestEditListboxForm.PrepareFirstShow;
var
  aPoint: TPoint;
begin
  FListBox.Items.Add(' ');
  aPoint := FSuggestEdit.ControlToScreen(Point(0, 0));
  SetBounds(aPoint.X, aPoint.Y + FSuggestEdit.Height, FSuggestEdit.Width, 10);
  FFirstShow := False;
end;

procedure TSuggestEditListboxForm.SelectItem;
begin
  if Assigned(FOnSelectItem) then
    FOnSelectItem(FListBox.GetSelectedText);
end;

procedure TSuggestEditListboxForm.SetColor(Value: TColor);
begin
  inherited SetColor(Value);
  if Assigned(FListBox) then
    FListBox.Color := Value;
end;

constructor TSuggestEditListboxForm.Create(AOwner: TComponent);
begin
  CreateNew(AOwner);
end;

constructor TSuggestEditListboxForm.CreateNew(AOwner: TComponent;
  Num: Integer);
begin
  inherited CreateNew(AOwner, Num);
  FSuggestEdit := TVirtualCustomSuggestEdit(Owner);
  BorderStyle := bsNone;
  FormStyle := fsStayOnTop;
  FListBox := TListBox.Create(Self);
  FListBox.Align := alClient;
  FListBox.Parent := Self;
  FListBox.OnSelectionChange := @SelectionChanged;
  FListBox.OnClick := @ClickItem;
  FListBox.OnMouseMove := @MouseMoveItems;
  FListBox.BorderStyle := bsSingle;
  FItemHeight := -1;
  FItemShowingCount := 0;
  FBorderHeight := -1;
  FItemsDummy := TStringList.Create;
  FItemsDummy.Sorted := True;
  FItemsDummy.Duplicates := dupIgnore;
  Items := TStringList.Create;
  Items.Sorted := True;
  Items.Duplicates := dupIgnore;
  FCurrentEditText := '';
  FFirstShow := True;
  OnActivate := @FocusEdit;
end;

destructor TSuggestEditListboxForm.Destroy;
begin
  FItemsDummy.Free;
  Items.Free;
  inherited Destroy;
end;

procedure TSuggestEditListboxForm.DeleteIndex;
var
  aIndex: Integer;
begin
  aIndex := FListBox.ItemIndex;
  FItemsDummy.Delete(aIndex);
  if FItemsDummy.Count = 0 then begin
    Visible := False;
    Exit;
  end;
  if aIndex >= FItemsDummy.Count then aIndex := Pred(FItemsDummy.Count);
  FillBox(FCurrentEditText);
  FListBox.ItemIndex := aIndex;
end;

procedure TSuggestEditListboxForm.FillBox(const s: String);
begin
  FCurrentEditText := s;
  FillItemsDummy;
  if FItemsDummy.Count = 0 then
    Visible := False
  else
  begin
    if not Showing then
    begin
      if FFirstShow then PrepareFirstShow;
      FListBox.ItemIndex := -1;
      Visible := True;
    end;
    ResizeForm;
    FListBox.Items.Assign(FItemsDummy);
  end;
end;

procedure TSuggestEditListboxForm.FocusOwner(Data: PtrInt);
begin
  if Assigned(FSuggestEdit) then FSuggestEdit.DoFocus;
end;

function TSuggestEditListboxForm.GetSelectedText: String;
begin
  Result := FListBox.GetSelectedText;
end;

procedure TSuggestEditListboxForm.MoveItemIndexDown;
begin
  if FListBox.ItemIndex < FListBox.Items.Count - 1 then
    FListBox.ItemIndex := FListBox.ItemIndex + 1;
end;

procedure TSuggestEditListboxForm.MoveItemIndexPageDown;
var
  aIndex: Integer;
begin
  aIndex := FListBox.ItemIndex + FItemShowingCount;
  if aIndex > FListBox.Items.Count - 1 then aIndex := FListBox.Items.Count - 1;
  FListBox.ItemIndex := aIndex;
end;

procedure TSuggestEditListboxForm.MoveItemIndexPageUp;
var
  aIndex: Integer;
begin
  if FListBox.ItemIndex <= 0 then begin
    FListBox.ItemIndex := -1;
    Exit;
  end;
  aIndex := FListBox.ItemIndex - FItemShowingCount;
  if aIndex < 0 then aIndex := 0;
  FListBox.ItemIndex := aIndex;
end;

procedure TSuggestEditListboxForm.MoveItemIndexUp;
begin
  if FListBox.ItemIndex > -1 then
    FListBox.ItemIndex := FListBox.ItemIndex - 1;
end;

procedure TSuggestEditListboxForm.SelectionChanged(Sender: TObject; User: boolean);
begin
  if Assigned(FOnSelectionChange) then
    FOnSelectionChange(SelectedItem);
end;

procedure TSuggestEditListboxForm.ResizeForm;
var
  aHeight: Integer;
  aWidth: Integer;
  aPoint: TPoint;
begin
  if FItemHeight <= 0 then FItemHeight := FListBox.ItemHeight;
  if FItemHeight <= 0 then Exit;
  if FBorderHeight = -1 then FBorderHeight := GetSystemMetrics(SM_CYFRAME);
  if FMaxPreviewCount > 0 then
    FItemShowingCount := MinIntValue([FItemsDummy.Count, (FMaxBoxHeight - FBorderHeight) div FItemHeight, FMaxPreviewCount])
  else
    FItemShowingCount := MinIntValue([FItemsDummy.Count, (FMaxBoxHeight - FBorderHeight) div FItemHeight]);
  FItemShowingCount := Max(0, FItemShowingCount);
  aHeight := FItemShowingCount * FItemHeight + FBorderHeight;
  aPoint := FSuggestEdit.ControlToScreen(Point(0, 0));
  aPoint.y := aPoint.y + FSuggestEdit.Height;
  aWidth := FSuggestEdit.Width;
  if (aPoint.y + aHeight > Screen.Height) and (aPoint.y - aHeight - FSuggestEdit.Height > 0) then
    aPoint.y := aPoint.y - aHeight - FSuggestEdit.Height;
  SetBounds(aPoint.x, aPoint.y, aWidth, aHeight);
end;

end.

